# Copyright 2020 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from fractions import Fraction
from abc import ABCMeta, abstractmethod
from typing import Generator, Tuple, Union, Optional, cast, Iterator, Type, TYPE_CHECKING, Protocol, runtime_checkable

from ..constants import MAX_NANOSEC
from ..exceptions import TsValueError

from ._types import RationalTypes
from .timestamp import Timestamp, SupportsMediaTimestamp, mediatimestamp

__all__ = ["TimeRange", "SupportsMediaTimeRange", "mediatimerange"]


if TYPE_CHECKING:
    @runtime_checkable
    class SupportsMediaTimeRange (Protocol):
        def __mediatimerange__(self) -> "TimeRange":
            ...
else:
    class SupportsMediaTimeRange (metaclass=ABCMeta):
        """This is an abstract base class for any class that can be automagically converted into a TimeRange.

        To implement this simply implement the __mediatimerange__ magic method. No need to inherit from this
        class explicitly.
        """
        @classmethod
        def __subclasshook__(cls, subclass: Type) -> bool:
            if (
                issubclass(subclass, TimeRange) or
                hasattr(subclass, "__mediatimerange__") or
                issubclass(subclass, SupportsMediaTimestamp)
            ):
                return True
            else:
                return False

        @abstractmethod
        def __mediatimerange__(self) -> "TimeRange":
            ...


def mediatimerange(v: SupportsMediaTimeRange) -> "TimeRange":
    """This method can be called on any object which supports the __mediatimerange__ magic method
    and also on a TimeRange. It will always return a TimeRange or raise a ValueError.
    """
    if isinstance(v, TimeRange):
        return v
    elif hasattr(v, "__mediatimerange__"):
        return v.__mediatimerange__()
    elif isinstance(v, SupportsMediaTimestamp):
        return mediatimerange(mediatimestamp(v))
    else:
        raise ValueError("{!r} cannot be converted to a mediatimestamp.TimeRange".format(v))


class TimeRange (object):
    """A nanosecond immutable precision time range object"""

    class Inclusivity (int):
        def __and__(self, other: int) -> "TimeRange.Inclusivity":
            return TimeRange.Inclusivity(int(self) & int(other) & 0x3)

        def __or__(self, other: int) -> "TimeRange.Inclusivity":
            return TimeRange.Inclusivity((int(self) | int(other)) & 0x3)

        def __xor__(self, other: int) -> "TimeRange.Inclusivity":
            return TimeRange.Inclusivity((int(self) ^ int(other)) & 0x3)

        def __invert__(self) -> "TimeRange.Inclusivity":
            return TimeRange.Inclusivity((~int(self)) & 0x3)

    EXCLUSIVE = Inclusivity(0x0)
    INCLUDE_START = Inclusivity(0x1)
    INCLUDE_END = Inclusivity(0x2)
    INCLUSIVE = Inclusivity(0x3)

    class Rounding(int):
        pass

    ROUND_DOWN = Rounding(0)
    ROUND_NEAREST = Rounding(1)
    ROUND_UP = Rounding(2)
    ROUND_IN = Rounding(3)
    ROUND_OUT = Rounding(4)
    ROUND_START = Rounding(5)
    ROUND_END = Rounding(6)

    def __init__(self,
                 start: Optional[SupportsMediaTimestamp],
                 end: Optional[SupportsMediaTimestamp],
                 inclusivity: "TimeRange.Inclusivity" = INCLUSIVE):
        """Construct a time range starting at start and ending at end

        :param start: A Timestamp or None
        :param end: A Timestamp or None
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END"""
        super().__init__()
        self.start: Optional[Timestamp]
        self.end: Optional[Timestamp]
        self.inclusivity: TimeRange.Inclusivity

        # Convert convertible inputs
        if start is not None:
            start = mediatimestamp(start)
        if end is not None:
            end = mediatimestamp(end)

        # Normalise the 'never' cases
        if start is not None and end is not None:
            if start > end or (start == end and inclusivity != TimeRange.INCLUSIVE):
                start = Timestamp()
                end = Timestamp()
                inclusivity = TimeRange.EXCLUSIVE

        # Normalise the 'eternity' cases
        if start is None and end is None:
            inclusivity = TimeRange.INCLUSIVE

        self.__dict__['start'] = start
        self.__dict__['end'] = end
        self.__dict__['inclusivity'] = inclusivity

    def __setattr__(self, name: str, value: object) -> None:
        raise TsValueError("Cannot assign to an immutable TimeRange")

    def __mediatimerange__(self) -> "TimeRange":
        return self

    def __iter__(self) -> Iterator[Timestamp]:
        return self.at_rate(MAX_NANOSEC)

    def __reversed__(self) -> Iterator[Timestamp]:
        return self.reversed_at_rate(MAX_NANOSEC)

    def __hash__(self) -> int:
        return hash(self.to_sec_nsec_range())

    def at_rate(self,
                numerator: RationalTypes,
                denominator: RationalTypes = 1,
                phase_offset: SupportsMediaTimestamp = Timestamp()) -> Iterator[Timestamp]:
        """Returns an iterable which yields Timestamp objects at the specified rate within the
        range starting at the beginning and moving later.

        :param numerator: The numerator for the rate in Hz (or the exact rate as a Fraction or float)
        :param denominator: The denominator for the rate in Hz
        :param phase_offset: A Timestamp object which sets the phase offset of the first timestamp
                             drawn from the iterable.

        :raises: ValueError If a phase_offset is specified which is larger than the reciprocal of the rate

        :returns: an iterable that yields Timestamp objects
        """
        rate = Fraction(numerator, denominator)
        phase_offset = mediatimestamp(phase_offset)
        if phase_offset >= Timestamp.from_count(1, rate):
            raise ValueError("phase_offset of {} is too large for rate {}".format(phase_offset, rate))

        if self.start is None:
            raise ValueError("Cannot iterate over a timerange with no start")

        count = (self.start - phase_offset).to_count(rate)

        while True:
            ts = Timestamp.from_count(count, rate) + phase_offset
            count += 1

            if ts < self.start or ((self.inclusivity & TimeRange.INCLUDE_START) == 0 and ts == self.start):
                continue
            elif (self.end is not None and
                  (ts > self.end or ((self.inclusivity & TimeRange.INCLUDE_END) == 0 and ts == self.end))):
                break
            else:
                yield ts

    def reversed_at_rate(self,
                         numerator: RationalTypes,
                         denominator: RationalTypes = 1,
                         phase_offset: SupportsMediaTimestamp = Timestamp()) -> Iterator[Timestamp]:
        """Returns an iterable which yields Timestamp objects at the specified rate within the
        range starting at the end and moving earlier.

        :param numerator: The numerator for the rate in Hz (or the exact rate as a Fraction or float)
        :param denominator: The denominator for the rate in Hz
        :param phase_offset: A Timestamp object which sets the phase offset of the first timestamp
                             drawn from the iterable.

        :raises: ValueError If a phase_offset is specified which is larger than the reciprocal of the rate

        :returns: an iterable that yields Timestamp objects
        """
        phase_offset = mediatimestamp(phase_offset)
        rate = Fraction(numerator, denominator)
        if phase_offset >= Timestamp.from_count(1, rate):
            raise ValueError("phase_offset of {} is too large for rate {}".format(phase_offset, rate))

        if self.end is None:
            raise ValueError("Cannot reverse iterate over a timerange with no end")

        count = (self.end - phase_offset).to_count(rate)

        while True:
            ts = Timestamp.from_count(count, rate) + phase_offset
            count -= 1

            if ts > self.end or ((self.inclusivity & TimeRange.INCLUDE_END) == 0 and ts == self.end):
                continue
            elif (self.start is not None and
                  (ts < self.start or ((self.inclusivity & TimeRange.INCLUDE_START) == 0 and ts == self.start))):
                break
            else:
                yield ts

    @classmethod
    def from_timerange(cls, other: SupportsMediaTimeRange) -> "TimeRange":
        """Construct an immutable timerange from another timerange"""
        other = mediatimerange(other)
        return TimeRange(other.start,
                         other.end,
                         other.inclusivity)

    @classmethod
    def from_start(cls, start: SupportsMediaTimestamp, inclusivity: "TimeRange.Inclusivity" = INCLUSIVE) -> "TimeRange":
        """Construct a time range starting at start with no end

        :param start: A Timestamp
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END"""
        return cls(start, None, inclusivity)

    @classmethod
    def from_end(cls, end: SupportsMediaTimestamp, inclusivity: "TimeRange.Inclusivity" = INCLUSIVE) -> "TimeRange":
        """Construct a time range ending at end with no start

        :param end: A Timestamp
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END"""
        return cls(None, end, inclusivity)

    @classmethod
    def from_start_length(cls,
                          start: SupportsMediaTimestamp,
                          length: SupportsMediaTimestamp,
                          inclusivity: "TimeRange.Inclusivity" = INCLUSIVE) -> "TimeRange":
        """Construct a time range starting at start and ending at (start + length)

        :param start: A Timestamp
        :param length: A Timestamp, which must be non-negative
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END

        :raises: TsValueError if the length is negative"""
        length = mediatimestamp(length)
        start = mediatimestamp(start)
        if length < Timestamp():
            raise TsValueError("Length must be non-negative")
        return cls(start, start + length, inclusivity)

    @classmethod
    def eternity(cls) -> "TimeRange":
        """Return an unbounded time range covering all time"""
        return cls(None, None)

    @classmethod
    def never(cls) -> "TimeRange":
        """Return a time range covering no time"""
        return cls(Timestamp(), Timestamp(), TimeRange.EXCLUSIVE)

    @classmethod
    def from_single_timestamp(cls, ts: SupportsMediaTimestamp) -> "TimeRange":
        """Construct a time range containing only a single timestamp

        :param ts: A Timestamp"""
        ts = mediatimestamp(ts)
        return cls(ts, ts, TimeRange.INCLUSIVE)

    @classmethod
    def from_str(cls, s: str, inclusivity: "TimeRange.Inclusivity" = INCLUSIVE) -> "TimeRange":
        """Convert a string to a time range.

        Valid ranges are:
        [<ts>_<ts>]
        [<ts>_<ts>)
        (<ts>_<ts>]
        (<ts>_<ts>)
        [<ts>]
        <ts>_<ts>
        <ts>
        ()

        where <ts> is any valid string format for Timestamp.from_str() or an empty string.

        The meaning of these is relatively simple: [ indicates including the start time,
        ( indicates excluding it, ] indicates including the end time, and ) indicates excludint it.
        If brackets are ommitted entirely then this is taken as an inclusive range at both ends.
        Omitting a timestamp indicates that there is no bound on that end (ie. the range goes on forever),
        including only a single timestamp by itself indicates a range containing exactly that one timestamp.
        Finally the string "()" represents the empty range.

        :param s: The string to process
        """
        m = re.match(r'(\[|\()?([^_\)\]]+)?(_([^_\)\]]+)?)?(\]|\))?', s)

        if m is None:
            raise ValueError("{!r} is not a valid TimeRange".format(s))

        inc = TimeRange.INCLUSIVE
        if m.group(1) == "(":
            inc &= ~TimeRange.INCLUDE_START
        if m.group(5) == ")":
            inc &= ~TimeRange.INCLUDE_END

        start_str = m.group(2)
        end_str = m.group(4)

        start: Optional[Timestamp] = None
        end: Optional[Timestamp] = None
        if start_str is not None:
            start = Timestamp.from_str(start_str)
        if end_str is not None:
            end = Timestamp.from_str(end_str)

        if start is None and end is None:
            # Ie. we have no first or second timestamp
            if m.group(3) is not None:
                # ie. we have a '_' character
                return cls.eternity()
            else:
                # We have no '_' character, so the whole range is empty
                return cls.never()
        elif start is not None and end is None and m.group(3) is None:
            # timestamp of form <ts>
            return cls.from_single_timestamp(start)
        else:
            return cls(start, end, inc)

    @property
    def length(self) -> Union[Timestamp, float]:
        if self.end is None or self.start is None:
            return float("inf")
        return self.end - self.start

    def bounded_before(self) -> bool:
        return self.start is not None

    def bounded_after(self) -> bool:
        return self.end is not None

    def unbounded(self) -> bool:
        return self.start is None and self.end is None

    def includes_start(self) -> bool:
        return (self.inclusivity & TimeRange.INCLUDE_START) != 0

    def includes_end(self) -> bool:
        return (self.inclusivity & TimeRange.INCLUDE_END) != 0

    def finite(self) -> bool:
        return (self.start is not None and self.end is not None)

    def __contains__(self, ts: object) -> bool:
        """Returns true if the timestamp is within this range."""
        return ((isinstance(ts, SupportsMediaTimestamp)) and
                (self.start is None or mediatimestamp(ts) >= self.start) and
                (self.end is None or mediatimestamp(ts) <= self.end) and
                (not ((self.start is not None) and
                      (mediatimestamp(ts) == self.start) and
                      (self.inclusivity & TimeRange.INCLUDE_START == 0))) and
                (not ((self.end is not None) and
                      (mediatimestamp(ts) == self.end) and
                      (self.inclusivity & TimeRange.INCLUDE_END == 0))))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SupportsMediaTimeRange):
            return False

        other = mediatimerange(other)
        return (isinstance(other, SupportsMediaTimeRange) and
                ((self.is_empty() and other.is_empty()) or
                (((self.start is None and other.start is None) or
                  (self.start == other.start and
                   (self.inclusivity & TimeRange.INCLUDE_START) == (other.inclusivity & TimeRange.INCLUDE_START))) and
                 ((self.end is None and other.end is None) or
                  (self.end == other.end and
                   (self.inclusivity & TimeRange.INCLUDE_END) == (other.inclusivity & TimeRange.INCLUDE_END))))))

    def __repr__(self) -> str:
        return "{}.{}.from_str('{}')".format("mediatimestamp.immutable", type(self).__name__, self.to_sec_nsec_range())

    def __str__(self) -> str:
        return self.to_sec_nsec_range()

    def contains_subrange(self, tr: SupportsMediaTimeRange) -> bool:
        """Returns True if the timerange supplied lies entirely inside this timerange"""
        tr = mediatimerange(tr)
        return ((not self.is_empty()) and
                (tr.is_empty() or
                 (self.start is None or (tr.start is not None and self.start <= tr.start)) and
                 (self.end is None or (tr.end is not None and self.end >= tr.end)) and
                 (not ((self.start is not None) and
                       (tr.start is not None) and
                       (self.start == tr.start) and
                       (self.inclusivity & TimeRange.INCLUDE_START == 0) and
                       (tr.inclusivity & TimeRange.INCLUDE_START != 0))) and
                 (not ((self.end is not None) and
                       (tr.end is not None) and
                       (self.end == tr.end) and
                       (self.inclusivity & TimeRange.INCLUDE_END == 0) and
                       (tr.inclusivity & TimeRange.INCLUDE_END != 0)))))

    def to_sec_nsec_range(self, with_inclusivity_markers: bool = True) -> str:
        """Convert to [<seconds>:<nanoseconds>_<seconds>:<nanoseconds>] format,
        usually the opening and closing delimiters are set to [ or ] for inclusive and ( or ) for exclusive ranges.
        Unbounded ranges have no marker attached to them.

        :param with_inclusivity_markers: if set to False do not include parentheses/brackets"""
        if self.is_empty():
            if with_inclusivity_markers:
                return "()"
            else:
                return ""
        elif self.start is not None and self.end is not None and self.start == self.end:
            if with_inclusivity_markers:
                return "[" + self.start.to_tai_sec_nsec() + "]"
            else:
                return self.start.to_tai_sec_nsec()

        if with_inclusivity_markers:
            brackets = [("(", ")"), ("[", ")"), ("(", "]"), ("[", "]")][self.inclusivity]
        else:
            brackets = ("", "")

        return '_'.join([
            (brackets[0] + self.start.to_tai_sec_nsec()) if self.start is not None else '',
            (self.end.to_tai_sec_nsec() + brackets[1]) if self.end is not None else ''
            ])

    def intersect_with(self, tr: SupportsMediaTimeRange) -> "TimeRange":
        """Return a range which represents the intersection of this range with another"""
        tr = mediatimerange(tr)
        if self.is_empty() or tr.is_empty():
            return TimeRange.never()

        start = self.start
        if tr.start is not None and (self.start is None or self.start < tr.start):
            start = tr.start
        end = self.end
        if tr.end is not None and (self.end is None or self.end > tr.end):
            end = tr.end

        inclusivity = TimeRange.EXCLUSIVE
        if start is None or (start in self and start in tr):
            inclusivity |= TimeRange.INCLUDE_START
        if end is None or (end in self and end in tr):
            inclusivity |= TimeRange.INCLUDE_END

        if start is not None and end is not None and start > end:
            return TimeRange.never()

        return TimeRange(start, end, inclusivity)

    def starts_inside_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if the start of this timerange is located inside the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                ((self.bounded_before() and self.start in other and
                  (not (other.bounded_after() and self.start == other.end and not self.includes_start()))) or
                 (self.bounded_before() and other.bounded_before() and self.start == other.start and
                  (not (self.includes_start() and not other.includes_start()))) or
                 (not self.bounded_before() and not other.bounded_before())))

    def ends_inside_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if the end of this timerange is located inside the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                ((self.bounded_after() and self.end in other and
                  (not (other.bounded_before() and self.end == other.start and not self.includes_end()))) or
                 (self.bounded_after() and other.bounded_after() and self.end == other.end and
                  (not (self.includes_end() and not other.includes_end()))) or
                 (not self.bounded_after() and not other.bounded_after())))

    def is_earlier_than_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if this timerange ends earlier than the start of the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_before() and
                self.bounded_after() and
                (cast(Timestamp, self.end) < cast(Timestamp, other.start) or
                 (cast(Timestamp, self.end) == cast(Timestamp, other.start) and
                  not (self.includes_end() and other.includes_start()))))

    def is_later_than_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if this timerange starts later than the end of the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_after() and
                self.bounded_before() and
                (cast(Timestamp, self.start) > cast(Timestamp, other.end) or
                 (cast(Timestamp, self.start) == cast(Timestamp, other.end) and
                  not (self.includes_start() and other.includes_end()))))

    def starts_earlier_than_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if this timerange starts earlier than the start of the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_before() and
                (not self.bounded_before() or
                 (cast(Timestamp, self.start) < cast(Timestamp, other.start) or
                  (cast(Timestamp, self.start) == cast(Timestamp, other.start) and
                   self.includes_start() and
                   not other.includes_start()))))

    def starts_later_than_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if this timerange starts later than the start of the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                self.bounded_before() and
                (not other.bounded_before() or
                 (cast(Timestamp, self.start) > cast(Timestamp, other.start) or
                  (cast(Timestamp, self.start) == cast(Timestamp, other.start) and
                   (not self.includes_start() and other.includes_start())))))

    def ends_earlier_than_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if this timerange ends earlier than the end of the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                self.bounded_after() and
                (not other.bounded_after() or
                 (cast(Timestamp, self.end) < cast(Timestamp, other.end) or
                  (cast(Timestamp, self.end) == cast(Timestamp, other.end) and
                   (not self.includes_end() and other.includes_end())))))

    def ends_later_than_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if this timerange ends later than the end of the other."""
        other = mediatimerange(other)
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_after() and
                (not self.bounded_after() or
                 (cast(Timestamp, self.end) > cast(Timestamp, other.end) or
                  (cast(Timestamp, self.end) == cast(Timestamp, other.end) and
                   self.includes_end() and
                   not other.includes_end()))))

    def overlaps_with_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if this timerange and the other overlap."""
        other = mediatimerange(other)
        return (not self.is_earlier_than_timerange(other) and not self.is_later_than_timerange(other))

    def is_contiguous_with_timerange(self, other: SupportsMediaTimeRange) -> bool:
        """Returns true if the union of this timerange and the other would be a valid timerange"""
        other = mediatimerange(other)
        return (self.overlaps_with_timerange(other) or
                (self.is_earlier_than_timerange(other) and
                 self.end == other.start and
                 (self.includes_end() or other.includes_start())) or
                (self.is_later_than_timerange(other) and
                 self.start == other.end and
                 (self.includes_start() or other.includes_end())))

    def union_with_timerange(self, other: SupportsMediaTimeRange) -> "TimeRange":
        """Returns the union of this timerange and the other.
        :raises: ValueError if the ranges are not contiguous."""
        other = mediatimerange(other)
        if not self.is_contiguous_with_timerange(other):
            raise ValueError("Timeranges {} and {} are not contiguous, so cannot take the union.".format(self, other))

        return self.extend_to_encompass_timerange(other)

    def extend_to_encompass_timerange(self, other: SupportsMediaTimeRange) -> "TimeRange":
        """Returns the timerange that encompasses this and the other timerange."""
        other = mediatimerange(other)
        if self.is_empty():
            return other

        if other.is_empty():
            return self

        inclusivity = TimeRange.EXCLUSIVE
        if self.start == other.start:
            start = self.start
            inclusivity |= ((self.inclusivity | other.inclusivity) & TimeRange.INCLUDE_START)
        elif self.starts_earlier_than_timerange(other):
            start = self.start
            inclusivity |= (self.inclusivity & TimeRange.INCLUDE_START)
        else:
            start = other.start
            inclusivity |= (other.inclusivity & TimeRange.INCLUDE_START)

        if self.end == other.end:
            end = self.end
            inclusivity |= ((self.inclusivity | other.inclusivity) & TimeRange.INCLUDE_END)
        elif self.ends_later_than_timerange(other):
            end = self.end
            inclusivity |= (self.inclusivity & TimeRange.INCLUDE_END)
        else:
            end = other.end
            inclusivity |= (other.inclusivity & TimeRange.INCLUDE_END)

        return TimeRange(start, end, inclusivity)

    def split_at(self, timestamp: SupportsMediaTimestamp) -> Tuple["TimeRange", "TimeRange"]:
        """Splits a timerange at a specified timestamp.

        It is guaranteed that the splitting point will be in the *second* TimeRange returned, and not in the first.

        :param timestamp: the timestamp to split at
        :returns: A pair of TimeRange objects
        :raises: ValueError if timestamp not in self"""

        timestamp = mediatimestamp(timestamp)

        if timestamp not in self:
            raise ValueError("Cannot split range {} at {}".format(self, timestamp))

        return (TimeRange(self.start, timestamp, (self.inclusivity & TimeRange.INCLUDE_START)),
                TimeRange(timestamp, self.end, TimeRange.INCLUDE_START | (self.inclusivity & TimeRange.INCLUDE_END)))

    def timerange_between(self, other: SupportsMediaTimeRange) -> "TimeRange":
        """Returns the time range between the end of the earlier timerange and the start of the later one"""
        other = mediatimerange(other)

        if self.is_contiguous_with_timerange(other):
            return TimeRange.never()
        elif self.is_earlier_than_timerange(other):
            inclusivity = TimeRange.EXCLUSIVE
            if not self.includes_end():
                inclusivity |= TimeRange.INCLUDE_START
            if not other.includes_start():
                inclusivity |= TimeRange.INCLUDE_END
            return TimeRange(self.end, other.start, inclusivity)
        else:
            inclusivity = TimeRange.EXCLUSIVE
            if not self.includes_start():
                inclusivity |= TimeRange.INCLUDE_END
            if not other.includes_end():
                inclusivity |= TimeRange.INCLUDE_START
            return TimeRange(other.end, self.start, inclusivity)

    def timerange_before(self) -> "TimeRange":
        """Returns the time range before the start of the this one"""
        if self.start is None:
            return TimeRange.never()

        if self.includes_start():
            inclusivity = TimeRange.EXCLUSIVE
        else:
            inclusivity = TimeRange.INCLUDE_END

        return TimeRange.from_end(self.start, inclusivity=inclusivity)

    def timerange_after(self) -> "TimeRange":
        """Returns the time range after the end of the this one"""
        if self.end is None:
            return TimeRange.never()

        if self.includes_end():
            inclusivity = TimeRange.EXCLUSIVE
        else:
            inclusivity = TimeRange.INCLUDE_START

        return TimeRange.from_start(self.end, inclusivity=inclusivity)

    def is_empty(self) -> bool:
        """Returns true on any empty range."""
        return (self.start is not None and
                self.end is not None and
                self.start == self.end and
                self.inclusivity != TimeRange.INCLUSIVE)

    def is_normalised(self,
                      rate_num: RationalTypes,
                      rate_den: RationalTypes = 1,
                      rounding: Rounding = ROUND_NEAREST) -> bool:
        """Checks if timerange is normalised"""
        normalised_timerange = self.normalise(rate_num, rate_den, rounding)
        if normalised_timerange == self:
            return True
        else:
            return False

    def normalise(self,
                  rate_num: RationalTypes,
                  rate_den: RationalTypes = 1,
                  rounding: Rounding = ROUND_NEAREST) -> "TimeRange":
        """Returns a normalised half-open TimeRange based on this timerange.

        The returned TimeRange will always have INCLUDE_START inclusivity.

        If the original TimeRange was inclusive of its start then the returned TimeRange will
        start at the normalised timestamp closest to that start point (respecting rounding).

        If the original TimeRange was exclusive of its start then the returned TimeRange will
        start at the next normalised timestamp after the normalised timestamp closest to that
        start point (respecting rounding).

        If the original TimeRange was exclusive of its end then the returned TimeRange will
        end just before the normalised timestamp closest to that end point (respecting rounding).

        If the original TimeRange was inclusive of its end then the returned TimeRange will
        end just before the next normalised timestamp after the normalised timestamp closest to that
        end point (respecting rounding).

        The rounding options are:
        * ROUND_NEAREST -- each end of the range independently rounds to the nearest normalised timestamp
        * ROUND_UP -- both ends of the range round up
        * ROUND_DOWN -- both ends of the range round down
        * ROUND_IN -- The start of the range rounds up, the end rounds down
        * ROUND_OUT -- The start of the range rounds down, the end rounds up
        * ROUND_START -- The start rounds to the nearest normalised timestamp, the end rounds in the same direction
                         as the start
        * ROUND_END -- The end rounds to the nearest normalised timestamp, the start rounds in the same direction
                       as the end
        """
        if rounding == TimeRange.ROUND_OUT:
            start_rounding = Timestamp.ROUND_DOWN
            end_rounding = Timestamp.ROUND_UP
        elif rounding == TimeRange.ROUND_IN:
            start_rounding = Timestamp.ROUND_UP
            end_rounding = Timestamp.ROUND_DOWN
        elif rounding in [TimeRange.ROUND_START, TimeRange.ROUND_END]:
            start_rounding = Timestamp.ROUND_NEAREST
            end_rounding = Timestamp.ROUND_NEAREST
        else:
            start_rounding = Timestamp.Rounding(rounding)
            end_rounding = Timestamp.Rounding(rounding)

        rate = Fraction(rate_num, rate_den)

        start: Optional[int]
        if self.bounded_before():
            start = cast(Timestamp, self.start).to_count(rate, rounding=start_rounding)
        else:
            start = None

        end: Optional[int]
        if self.bounded_after():
            end = cast(Timestamp, self.end).to_count(rate, rounding=end_rounding)
        else:
            end = None

        if rounding == TimeRange.ROUND_START and self.bounded_before() and self.bounded_after():
            if start == cast(Timestamp, self.start).to_count(rate, rounding=Timestamp.ROUND_UP):
                end = cast(Timestamp, self.end).to_count(rate, rounding=Timestamp.ROUND_UP)
            else:
                end = cast(Timestamp, self.end).to_count(rate, rounding=Timestamp.ROUND_DOWN)
        elif rounding == TimeRange.ROUND_END and self.bounded_before() and self.bounded_after():
            if end == cast(Timestamp, self.end).to_count(rate, rounding=Timestamp.ROUND_UP):
                start = cast(Timestamp, self.start).to_count(rate, rounding=Timestamp.ROUND_UP)
            else:
                start = cast(Timestamp, self.start).to_count(rate, rounding=Timestamp.ROUND_DOWN)

        if start is not None and not self.includes_start():
            start += 1
        if end is not None and self.includes_end():
            end += 1

        start_ts: Optional[Timestamp] = None
        end_ts: Optional[Timestamp] = None
        if start is not None:
            start_ts = Timestamp.from_count(start, rate)
        if end is not None:
            end_ts = Timestamp.from_count(end, rate)

        return TimeRange(start_ts,
                         end_ts,
                         TimeRange.INCLUDE_START)

    def into_chunks(
        self,
        time_duration: Timestamp
    ) -> Generator["TimeRange", None, None]:
        """Returns a generator of TimeRanges of the length specified in time_duration based on this TimeRange.

        If this TimeRange has an infinite length, then this will generate chunks of time_duration infinitely.

        The first chunk will have the start inclusivity of this TimeRange, and will have an exclusive end.
        The last chunk (if this TimeRange is not infinite) will have the end inclusivity of this TimeRange, and an
        inclusive start.
        All other chunks will have an inclusive start and exclusive end.

        If the time_duration does not divide exactly into the length of this TimeRange, the last chunk returned will
        be of the 'remainder' length.

        If the time_duration is the same length or longer than this TimeRange, then a copy of this TimeRange will be
        returned.

        :param time_duration: A Timestamp representing the requested chunk length
        :returns: A TimeRange generator that generates chunks.
        """
        if not self.bounded_before():
            raise ValueError("The timerange to be chunked has no start time!")
        remainder = self  # this is a timerange from the end of the last one to the end of self
        while not remainder.is_empty():
            next_timerange = TimeRange.from_start_length(
                cast(Timestamp, remainder.start),
                time_duration,
                TimeRange.INCLUDE_START)
            next_timerange = next_timerange.intersect_with(remainder)

            if next_timerange.end == remainder.end and remainder.includes_end():
                next_timerange = TimeRange.from_start_length(
                    cast(Timestamp, next_timerange.start),
                    cast(Timestamp, next_timerange.length),
                    next_timerange.inclusivity | TimeRange.INCLUDE_END
                )

            yield next_timerange
            # If we have just yielded the final chunk, then this will produce an empty timerange.
            remainder = remainder.intersect_with(next_timerange.timerange_after())
        return



# Copyright 2019 British Broadcasting Corporation
#
# This is an internal BBC tool and is not licensed externally
# If you have received a copy of this erroneously then you do
# not have permission to reproduce it.

import unittest

from mediatimestamp import CountRange


class TestCountRange(unittest.TestCase):
    def test_never(self):
        rng = CountRange.never()

        self.assertNotIn(-1, rng)
        self.assertNotIn(0, rng)
        self.assertNotIn(1, rng)

        self.assertTrue(rng.is_empty())
        self.assertEqual(rng.to_str(), "()")

    def test_eternity(self):
        alltime = CountRange.eternity()

        self.assertIn(-1, alltime)
        self.assertIn(0, alltime)
        self.assertIn(1, alltime)

        self.assertEqual(alltime.to_str(), "_")

    def test_bounded_on_right_inclusive(self):
        rng = CountRange.from_end(100)

        self.assertIn(-1, rng)
        self.assertIn(0, rng)
        self.assertIn(99, rng)
        self.assertIn(100, rng)
        self.assertNotIn(101, rng)

        self.assertEqual(rng.to_str(), "_101)")

    def test_bounded_on_right_exclusive(self):
        rng = CountRange.from_end(100, inclusivity=CountRange.EXCLUSIVE)

        self.assertIn(-1, rng)
        self.assertIn(0, rng)
        self.assertIn(99, rng)
        self.assertNotIn(100, rng)
        self.assertNotIn(101, rng)

        self.assertEqual(rng.to_str(), "_100)")

    def test_bounded_on_left_inclusive(self):
        rng = CountRange.from_start(100)

        self.assertNotIn(-1, rng)
        self.assertNotIn(0, rng)
        self.assertNotIn(99, rng)
        self.assertIn(100, rng)
        self.assertIn(101, rng)

        self.assertEqual(rng.to_str(), "[100_")

    def test_bounded_on_left_exclusive(self):
        rng = CountRange.from_start(100, inclusivity=CountRange.EXCLUSIVE)

        self.assertNotIn(-1, rng)
        self.assertNotIn(0, rng)
        self.assertNotIn(99, rng)
        self.assertNotIn(100, rng)
        self.assertIn(101, rng)

        self.assertEqual(rng.to_str(), "[101_")

    def test_bounded_inclusive(self):
        rng = CountRange(100, 200)

        self.assertNotIn(-1, rng)
        self.assertNotIn(0, rng)
        self.assertNotIn(99, rng)
        self.assertIn(100, rng)
        self.assertIn(101, rng)
        self.assertIn(199, rng)
        self.assertIn(200, rng)
        self.assertNotIn(201, rng)

        self.assertEqual(rng.to_str(), "[100_201)")

    def test_bounded_exclusive(self):
        rng = CountRange(100, 200, inclusivity=CountRange.EXCLUSIVE)

        self.assertNotIn(-1, rng)
        self.assertNotIn(0, rng)
        self.assertNotIn(99, rng)
        self.assertNotIn(100, rng)
        self.assertIn(101, rng)
        self.assertIn(199, rng)
        self.assertNotIn(200, rng)
        self.assertNotIn(201, rng)

        self.assertEqual(rng.to_str(), "[101_200)")

    def test_single_tv(self):
        rng = CountRange.from_single_count(100)

        self.assertNotIn(-1, rng)
        self.assertNotIn(0, rng)
        self.assertNotIn(99, rng)
        self.assertIn(100, rng)
        self.assertNotIn(101, rng)

        self.assertEqual(rng.to_str(), "[100_101)")

    def test_from_str(self):
        tests = [
            ("()", CountRange.never()),
            ("[]", CountRange.never()),
            ("", CountRange.never()),
            ("_", CountRange.eternity()),
            ("_200", CountRange.from_end(200)),
            ("[_200]", CountRange.from_end(200, inclusivity=CountRange.INCLUSIVE)),
            ("(_200)", CountRange.from_end(200, inclusivity=CountRange.EXCLUSIVE)),
            ("100_", CountRange.from_start(100)),
            ("[100_]", CountRange.from_start(100, inclusivity=CountRange.INCLUSIVE)),
            ("(100_)", CountRange.from_start(100, inclusivity=CountRange.EXCLUSIVE)),
            ("100_200", CountRange(100, 200)),
            ("[100_200]", CountRange(100, 200, inclusivity=CountRange.INCLUSIVE)),
            ("(100_200)", CountRange(100, 200, inclusivity=CountRange.EXCLUSIVE)),
            ("(100_200]", CountRange(100, 200, inclusivity=CountRange.INCLUDE_END)),
            ("[100_200)", CountRange(100, 200, inclusivity=CountRange.INCLUDE_START)),
            ("200", CountRange.from_single_count(200)),
        ]

        for (s, tr) in tests:
            self.assertEqual(tr, CountRange.from_str(s))

    def test_subrange(self):
        a = 50
        b = 100
        c = 200
        d = 250

        self.assertTrue(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(a, b, inclusivity=CountRange.INCLUSIVE)))
        self.assertFalse(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(a, b, inclusivity=CountRange.INCLUSIVE)))
        self.assertTrue(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(a, b, inclusivity=CountRange.EXCLUSIVE)))
        self.assertTrue(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(a, b, inclusivity=CountRange.EXCLUSIVE)))

        self.assertTrue(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)))
        self.assertFalse(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)))
        self.assertTrue(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.EXCLUSIVE)))
        self.assertTrue(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.EXCLUSIVE)))

        self.assertTrue(CountRange(a, d, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)))
        self.assertTrue(CountRange(a, d, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)))
        self.assertTrue(CountRange(a, d, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.EXCLUSIVE)))
        self.assertTrue(CountRange(a, d, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(b, c, inclusivity=CountRange.EXCLUSIVE)))

        self.assertFalse(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(b, d, inclusivity=CountRange.INCLUSIVE)))
        self.assertFalse(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(b, d, inclusivity=CountRange.INCLUSIVE)))
        self.assertFalse(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).contains_subrange(
            CountRange(b, d, inclusivity=CountRange.EXCLUSIVE)))
        self.assertFalse(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).contains_subrange(
            CountRange(b, d, inclusivity=CountRange.EXCLUSIVE)))

    def test_intersection(self):
        a = 50
        b = 100
        c = 200
        d = 250

        self.assertEqual(CountRange(a, b, inclusivity=CountRange.INCLUSIVE).intersect_with(
                            CountRange(c, d, inclusivity=CountRange.INCLUSIVE)),
                         CountRange.never())

        self.assertEqual(CountRange(a, b, inclusivity=CountRange.INCLUSIVE).intersect_with(
                            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)),
                         CountRange.from_single_count(b))

        self.assertEqual(CountRange(a, b, inclusivity=CountRange.EXCLUSIVE).intersect_with(
                            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)),
                         CountRange.never())

        self.assertEqual(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).intersect_with(
                            CountRange(b, d, inclusivity=CountRange.INCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.INCLUSIVE))

        self.assertEqual(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).intersect_with(
                            CountRange(b, d, inclusivity=CountRange.INCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.INCLUDE_START))

        self.assertEqual(CountRange(a, c, inclusivity=CountRange.INCLUSIVE).intersect_with(
                            CountRange(b, d, inclusivity=CountRange.EXCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.INCLUDE_END))

        self.assertEqual(CountRange(a, c, inclusivity=CountRange.EXCLUSIVE).intersect_with(
                            CountRange(b, d, inclusivity=CountRange.EXCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.EXCLUSIVE))

        self.assertEqual(CountRange(a, d, inclusivity=CountRange.INCLUSIVE).intersect_with(
                            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.INCLUSIVE))

        self.assertEqual(CountRange(a, d, inclusivity=CountRange.EXCLUSIVE).intersect_with(
                            CountRange(b, c, inclusivity=CountRange.INCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.INCLUSIVE))

        self.assertEqual(CountRange(a, d, inclusivity=CountRange.INCLUSIVE).intersect_with(
                            CountRange(b, c, inclusivity=CountRange.EXCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.EXCLUSIVE))

        self.assertEqual(CountRange(a, d, inclusivity=CountRange.EXCLUSIVE).intersect_with(
                            CountRange(b, c, inclusivity=CountRange.EXCLUSIVE)),
                         CountRange(b, c, inclusivity=CountRange.EXCLUSIVE))

        self.assertEqual(CountRange.eternity().intersect_with(
                            CountRange(a, b, inclusivity=CountRange.INCLUSIVE)),
                         CountRange(a, b, inclusivity=CountRange.INCLUSIVE))

        self.assertEqual(CountRange.eternity().intersect_with(
                            CountRange(a, b, inclusivity=CountRange.EXCLUSIVE)),
                         CountRange(a, b, inclusivity=CountRange.EXCLUSIVE))

        self.assertEqual(CountRange.never().intersect_with(
                            CountRange(a, b, inclusivity=CountRange.INCLUSIVE)),
                         CountRange.never())

        self.assertEqual(CountRange.never().intersect_with(
                            CountRange(a, b, inclusivity=CountRange.EXCLUSIVE)),
                         CountRange.never())

        self.assertEqual(CountRange.never().intersect_with(
                            CountRange.eternity()),
                         CountRange.never())

    def test_length(self):
        a = 50
        b = 100
        c = 200

        rng = CountRange(a, b, inclusivity=CountRange.INCLUSIVE)
        self.assertEqual(rng.length, b - a + 1)

        rng = CountRange(None, b, inclusivity=CountRange.INCLUSIVE)
        self.assertEqual(rng.length, float("inf"))

        rng = CountRange(a, None, inclusivity=CountRange.INCLUSIVE)
        self.assertEqual(rng.length, float("inf"))

        rng = CountRange(None, None, inclusivity=CountRange.INCLUSIVE)
        self.assertEqual(rng.length, float("inf"))

        rng = CountRange(a, b, inclusivity=CountRange.INCLUSIVE)
        with self.assertRaises(ValueError):
            rng.length = (c - a)
        self.assertEqual(rng, CountRange(a, b, inclusivity=CountRange.INCLUSIVE))

        rng = CountRange(None, None, inclusivity=CountRange.INCLUSIVE)
        with self.assertRaises(ValueError):
            rng.length = (b - a)

        rng = CountRange(a, b, inclusivity=CountRange.INCLUSIVE)
        with self.assertRaises(ValueError):
            rng.length = (a - c)

    def test_repr(self):
        """This tests that the repr function turns time ranges into `eval`-able strings."""
        test_trs = [
            (CountRange.from_str("(100_200)"), "mediatimestamp.count_range.CountRange.from_str('[101_200)')"),
            (CountRange.from_str("[100_200]"), "mediatimestamp.count_range.CountRange.from_str('[100_201)')"),
            (CountRange.from_str("[100_"), "mediatimestamp.count_range.CountRange.from_str('[100_')")
        ]

        for t in test_trs:
            self.assertEqual(repr(t[0]), t[1])

    def test_comparisons(self):
        # Test data format:
        #  (a, b,
        #   (starts_inside, ends_inside, is_earlier, is_later,
        #    starts_earlier, starts_later, ends_earlier, ends_later,
        #    overlaps_with, is_contiguous_with))
        test_data = [
            (CountRange.from_str("_"), CountRange.from_str("_"),
             (True, True, False, False, False, False, False, False, True, True)),
            (CountRange.from_str("_"), CountRange.from_str("[0_"),
             (False, True, False, False, True, False, False, False, True, True)),
            (CountRange.from_str("_"), CountRange.from_str("_0]"),
             (True, False, False, False, False, False, False, True, True, True)),
            (CountRange.from_str("_"), CountRange.from_str("[0_10)"),
             (False, False, False, False, True, False, False, True, True, True)),

            (CountRange.from_str("_5)"), CountRange.from_str("_"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("_5)"), CountRange.from_str("[0_"),
             (False, True, False, False, True, False, True, False, True, True)),
            (CountRange.from_str("_5)"), CountRange.from_str("_0]"),
             (True, False, False, False, False, False, False, True, True, True)),
            (CountRange.from_str("_5)"), CountRange.from_str("_10]"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("_5)"), CountRange.from_str("[0_10)"),
             (False, True, False, False, True, False, True, False, True, True)),

            (CountRange.from_str("_0)"), CountRange.from_str("_"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("_0)"), CountRange.from_str("[0_"),
             (False, False, True, False, True, False, True, False, False, True)),
            (CountRange.from_str("_0)"), CountRange.from_str("_0]"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("_0)"), CountRange.from_str("_0)"),
             (True, True, False, False, False, False, False, False, True, True)),
            (CountRange.from_str("_0)"), CountRange.from_str("_10]"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("_0)"), CountRange.from_str("[0_10)"),
             (False, False, True, False, True, False, True, False, False, True)),
            (CountRange.from_str("_0)"), CountRange.from_str("(0_10)"),
             (False, False, True, False, True, False, True, False, False, False)),
            (CountRange.from_str("_0)"), CountRange.from_str("[5_10)"),
             (False, False, True, False, True, False, True, False, False, False)),

            (CountRange.from_str("[0_)"), CountRange.from_str("_"),
             (True, True, False, False, False, True, False, False, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("[0_"),
             (True, True, False, False, False, False, False, False, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("(0_"),
             (False, True, False, False, True, False, False, False, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("[5_"),
             (False, True, False, False, True, False, False, False, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("_0]"),
             (True, False, False, False, False, True, False, True, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("_0)"),
             (False, False, False, True, False, True, False, True, False, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("_10]"),
             (True, False, False, False, False, True, False, True, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("[0_10)"),
             (True, False, False, False, False, False, False, True, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("(0_10)"),
             (False, False, False, False, True, False, False, True, True, True)),
            (CountRange.from_str("[0_)"), CountRange.from_str("[5_10)"),
             (False, False, False, False, True, False, False, True, True, True)),

            (CountRange.from_str("[5_)"), CountRange.from_str("_"),
             (True, True, False, False, False, True, False, False, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("[0_"),
             (True, True, False, False, False, True, False, False, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("(0_"),
             (True, True, False, False, False, True, False, False, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("[5_"),
             (True, True, False, False, False, False, False, False, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("(5_"),
             (False, True, False, False, True, False, False, False, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("_0]"),
             (False, False, False, True, False, True, False, True, False, False)),
            (CountRange.from_str("[5_)"), CountRange.from_str("_0)"),
             (False, False, False, True, False, True, False, True, False, False)),
            (CountRange.from_str("[5_)"), CountRange.from_str("_10]"),
             (True, False, False, False, False, True, False, True, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("[0_10)"),
             (True, False, False, False, False, True, False, True, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("(0_10)"),
             (True, False, False, False, False, True, False, True, True, True)),
            (CountRange.from_str("[5_)"), CountRange.from_str("[5_10)"),
             (True, False, False, False, False, False, False, True, True, True)),

            (CountRange.from_str("[0_10)"), CountRange.from_str("_"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[0_"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("(0_"),
             (False, True, False, False, True, False, True, False, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[10_"),
             (False, False, True, False, True, False, True, False, False, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("(10_"),
             (False, False, True, False, True, False, True, False, False, False)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("_0]"),
             (True, False, False, False, False, True, False, True, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("_0)"),
             (False, False, False, True, False, True, False, True, False, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("_10]"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("_10)"),
             (True, True, False, False, False, True, False, False, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[0_10)"),
             (True, True, False, False, False, False, False, False, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("(0_10)"),
             (False, True, False, False, True, False, False, False, True, True)),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[5_10)"),
             (False, True, False, False, True, False, False, False, True, True)),

            (CountRange.from_str("(0_10)"), CountRange.from_str("_"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("[0_"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("(0_"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("[10_"),
             (False, False, True, False, True, False, True, False, False, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("(10_"),
             (False, False, True, False, True, False, True, False, False, False)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("_0]"),
             (False, False, False, True, False, True, False, True, False, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("_0)"),
             (False, False, False, True, False, True, False, True, False, False)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("_10]"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("_10)"),
             (True, True, False, False, False, True, False, False, True, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("[0_10)"),
             (True, True, False, False, False, True, False, False, True, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("(0_10)"),
             (True, True, False, False, False, False, False, False, True, True)),
            (CountRange.from_str("(0_10)"), CountRange.from_str("[5_10)"),
             (False, True, False, False, True, False, False, False, True, True)),

            (CountRange.from_str("[0_5)"), CountRange.from_str("_"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("[0_"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("(0_"),
             (False, True, False, False, True, False, True, False, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("[10_"),
             (False, False, True, False, True, False, True, False, False, False)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("(10_"),
             (False, False, True, False, True, False, True, False, False, False)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("_0]"),
             (True, False, False, False, False, True, False, True, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("_0)"),
             (False, False, False, True, False, True, False, True, False, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("_10]"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("_10)"),
             (True, True, False, False, False, True, True, False, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("[0_10)"),
             (True, True, False, False, False, False, True, False, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("(0_10)"),
             (False, True, False, False, True, False, True, False, True, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("[5_10)"),
             (False, False, True, False, True, False, True, False, False, True)),
            (CountRange.from_str("[0_5)"), CountRange.from_str("(5_10)"),
             (False, False, True, False, True, False, True, False, False, False)),

            (CountRange.never(), CountRange.from_str("_"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("[0_"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("(0_"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("[10_"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("(10_"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("_0]"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("_0)"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("_10]"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("_10)"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("[0_10)"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("(0_10)"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("[5_10)"),
             (False, False, False, False, False, False, False, False, True, True)),
            (CountRange.never(), CountRange.from_str("(5_10)"),
             (False, False, False, False, False, False, False, False, True, True)),
        ]
        functions = ("starts_inside_range",
                     "ends_inside_range",
                     "is_earlier_than_range",
                     "is_later_than_range",
                     "starts_earlier_than_range",
                     "starts_later_than_range",
                     "ends_earlier_than_range",
                     "ends_later_than_range",
                     "overlaps_with_range",
                     "is_contiguous_with_range")

        for (a, b, expected) in test_data:
            for (fname, expected_value) in zip(functions, expected):
                with self.subTest(a=a, b=b, fname=fname, expected_value=expected_value):
                    if expected_value:
                        self.assertTrue(getattr(a, fname)(b),
                                        msg="{!r}.{}({!r}) is False, expected to be True".format(a, fname, b))
                    else:
                        self.assertFalse(getattr(a, fname)(b),
                                         msg="{!r}.{}({!r}) is True, expected to be False".format(a, fname, b))

    def test_split(self):
        test_data = [
            (CountRange.from_str("_"), 0,
             CountRange.from_str("_0)"), CountRange.from_str("[0_")),
            (CountRange.from_str("[0_"), 10,
             CountRange.from_str("[0_10)"), CountRange.from_str("[10_")),
            (CountRange.from_str("_10)"), 0,
             CountRange.from_str("_0)"), CountRange.from_str("[0_10)")),
            (CountRange.from_str("[0_10)"), 5,
             CountRange.from_str("[0_5)"), CountRange.from_str("[5_10)")),
            (CountRange.from_str("[0_10]"), 5,
             CountRange.from_str("[0_5)"), CountRange.from_str("[5_10]")),
            (CountRange.from_str("(0_10)"), 5,
             CountRange.from_str("(0_5)"), CountRange.from_str("[5_10)")),
            (CountRange.from_str("(0_10]"), 5,
             CountRange.from_str("(0_5)"), CountRange.from_str("[5_10]")),
            (CountRange.from_str("[0]"), 0,
             CountRange.never(), CountRange.from_str("[0_0]")),
            (CountRange.from_str("[0_10)"), 0,
             CountRange.never(), CountRange.from_str("[0_10)")),
            (CountRange.from_str("[0_10]"), 10,
             CountRange.from_str("[0_10)"), CountRange.from_str("[10]")),
        ]

        for (tr, ts, left, right) in test_data:
            with self.subTest(tr=tr, ts=ts, expected=(left, right)):
                self.assertEqual(tr.split_at(ts), (left, right))

        test_data = [
            (CountRange.from_str("[0_10)"), 11),
            (CountRange.from_str("[0_10)"), 10),
        ]

        for (tr, ts) in test_data:
            with self.subTest(tr=tr, ts=ts):
                with self.assertRaises(ValueError):
                    tr.split_at(ts)

    def test_extend_to_encompass(self):
        test_data = [
            (CountRange.from_str("()"), CountRange.from_str("()"),
             CountRange.from_str("()")),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[10]"),
             CountRange.from_str("[0_10]")),
            (CountRange.from_str("_"), CountRange.from_str("[0]"),
             CountRange.from_str("_")),
            (CountRange.from_str("_"), CountRange.from_str("()"),
             CountRange.from_str("_")),
            (CountRange.from_str("()"), CountRange.from_str("_"),
             CountRange.from_str("_")),
            (CountRange.from_str("_10)"), CountRange.from_str("[0_"),
             CountRange.from_str("_")),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[5_"),
             CountRange.from_str("[0_")),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[5_15)"),
             CountRange.from_str("[0_15)")),
            (CountRange.from_str("[0_10)"), CountRange.from_str("[10_15)"),
             CountRange.from_str("[0_15)")),
            (CountRange.from_str("()"), CountRange.from_str("[5_"),
             CountRange.from_str("[5_")),
            (CountRange.from_str("()"), CountRange.from_str("[5_15)"),
             CountRange.from_str("[5_15)")),
            (CountRange.from_str("()"), CountRange.from_str("_15)"),
             CountRange.from_str("_15)")),

            # discontiguous
            (CountRange.from_str("_0)"), CountRange.from_str("(0_"),
             CountRange.from_str("_")),
            (CountRange.from_str("(0_"), CountRange.from_str("_0)"),
             CountRange.from_str("_")),
            (CountRange.from_str("[0_5)"), CountRange.from_str("(5_15)"),
             CountRange.from_str("[0_15)")),
            (CountRange.from_str("(5_15)"), CountRange.from_str("[0_5)"),
             CountRange.from_str("[0_15)")),
            (CountRange.from_str("[0_5)"), CountRange.from_str("[10_15)"),
             CountRange.from_str("[0_15)")),
            (CountRange.from_str("[10_15)"), CountRange.from_str("[0_5)"),
             CountRange.from_str("[0_15)")),
        ]

        for (first, second, expected) in test_data:
            with self.subTest(first=first, second=second, expected=expected):
                self.assertEqual(first.extend_to_encompass_range(second), expected)

    def test_union_raises(self):
        # discontiguous part of test_extend_to_encompass raises for a union
        test_data = [
            (CountRange.from_str("_0)"), CountRange.from_str("(0_")),
            (CountRange.from_str("(0_"), CountRange.from_str("_0)")),
            (CountRange.from_str("[0_5)"), CountRange.from_str("(5_15)")),
            (CountRange.from_str("(5_15)"), CountRange.from_str("[0_5)")),
            (CountRange.from_str("[0_5)"), CountRange.from_str("[10_15)")),
            (CountRange.from_str("[10_15)"), CountRange.from_str("[0_5)")),
        ]

        for (first, second) in test_data:
            with self.subTest(first=first, second=second):
                with self.assertRaises(ValueError):
                    first.union_with_range(second)

    def test_immutable(self):
        cr = CountRange(0, 1)
        with self.assertRaises(ValueError):
            cr._start = 1

        with self.assertRaises(ValueError):
            cr._end = 1

        with self.assertRaises(ValueError):
            cr._inclusivity = CountRange.INCLUSIVE



# Copyright 2020 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This library provides a class Timestamp which stores an immutable signed time difference value with nanosecond
precision.

The Timestamp can represent an immutable time offset since the epoch (ie. 1970-01-01T00:00:00.000000000Z)

This library also includes an immutable TimeRange object which stores a range between two Timestamps.

These data types are of use in a number of situations, but particularly for code that will handle PTP timestamps, which
are normally stored in this fashion.
"""

from .timeoffset import TimeOffset, SupportsMediaTimeOffset, mediatimeoffset
from .timestamp import Timestamp, SupportsMediaTimestamp, mediatimestamp
from .timerange import TimeRange, SupportsMediaTimeRange, mediatimerange
from ..exceptions import TsValueError


__all__ = [
    "TimeOffset", "SupportsMediaTimeOffset", "mediatimeoffset",
    "Timestamp", "SupportsMediaTimestamp", "mediatimestamp",
    "TimeRange", "SupportsMediaTimeRange", "mediatimerange",
    "TsValueError"]



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Optional, Type, TYPE_CHECKING, Protocol, runtime_checkable, Union
from abc import ABCMeta, abstractmethod
import calendar
import time
import re
from datetime import datetime
from dateutil import tz
from fractions import Fraction

from deprecated import deprecated

from ..constants import MAX_NANOSEC, MAX_SECONDS, UTC_LEAP
from ..exceptions import TsValueError

from ._parse import _parse_seconds_fraction, _parse_iso8601
from ._types import RationalTypes

if TYPE_CHECKING:
    from .timerange import TimeRange  # noqa: F401

__all__ = ["Timestamp", "SupportsMediaTimestamp", "mediatimestamp", "TimestampConstructionType"]


TimestampConstructionType = Union["Timestamp", "SupportsMediaTimestamp", int, float]


if TYPE_CHECKING:
    @runtime_checkable
    class SupportsMediaTimestamp (Protocol):
        def __mediatimestamp__(self) -> "Timestamp":
            ...
else:
    class SupportsMediaTimestamp (metaclass=ABCMeta):
        """This is an abstract base class for any class that can be automagically converted into a Timestamp.

        To implement this simply implement the __mediatimestamp__ magic method. No need to inherit from this
        class explicitly.
        """
        @classmethod
        def __subclasshook__(cls, subclass: Type) -> bool:
            if hasattr(subclass, "__mediatimestamp__"):
                return True
            else:
                return False

        @abstractmethod
        def __mediatimestamp__(self) -> "Timestamp":
            ...


def mediatimestamp(v: TimestampConstructionType) -> "Timestamp":
    """This method can be called on any object which supports the __mediatimestamp__ magic method
    and also on a Timestamp, an int or a float.
    It will always return a Timestamp or raise a ValueError.
    """
    if isinstance(v, Timestamp):
        return v
    elif isinstance(v, int):
        return Timestamp(v)
    elif isinstance(v, float):
        return Timestamp.from_float(v)
    elif hasattr(v, "__mediatimestamp__"):
        return v.__mediatimestamp__()
    else:
        raise ValueError("{!r} cannot be converted to a mediatimestamp.Timestamp".format(v))


class Timestamp(object):
    """A nanosecond precision immutable timestamp.

    Note that the canonical representation of a Timestamp is seconds:nanoseconds, e.g. "4:500000000".
    Timestamp in seconds.fractions format (e.g. "4.5") can be parsed, but should not be used for serialization or
    storage due to difficulty disambiguating them from floats.
    """
    class Rounding (int):
        pass

    ROUND_DOWN = Rounding(0)
    ROUND_NEAREST = Rounding(1)
    ROUND_UP = Rounding(2)

    MAX_NANOSEC = MAX_NANOSEC
    MAX_SECONDS = MAX_SECONDS

    def __init__(self, sec: int = 0, ns: int = 0, sign: int = 1):
        if sign < 0:
            sign = -1
        else:
            sign = 1
        value = sign * int(sec * self.MAX_NANOSEC + ns)

        value_limit = self.MAX_SECONDS * self.MAX_NANOSEC - 1
        value = max(-value_limit, min(value_limit, value))

        self._value: int

        self.__dict__['_value'] = value

    @property
    def sec(self) -> int:
        """Returns the whole number of seconds"""
        return int(abs(self._value) // self.MAX_NANOSEC)

    @property
    def ns(self) -> int:
        """Returns the nanoseconds remainder after subtrating the whole number of seconds"""
        return abs(self._value) - self.sec * self.MAX_NANOSEC

    @property
    def sign(self) -> int:
        """Returns 1 if the timeoffset is positive, -1 if negative"""
        if self._value < 0:
            return -1
        else:
            return 1

    def __setattr__(self, name: str, value: object) -> None:
        raise TsValueError("Cannot assign to an immutable Timestamp")

    def __mediatimestamp__(self) -> "Timestamp":
        return self

    @deprecated(version="4.0.0",
                reason="This method is deprecated. TimeOffset has been merged into Timestamp.")
    def __mediatimeoffset__(self) -> "Timestamp":
        """Legacy method for getting a TimeOffset"""
        return self

    def __mediatimerange__(self) -> "TimeRange":
        from .timerange import TimeRange  # noqa: F811

        return TimeRange.from_single_timestamp(self)

    @classmethod
    def get_time(cls) -> "Timestamp":
        unix_time = time.time()
        return cls.from_unix(int(unix_time), int(unix_time*cls.MAX_NANOSEC) - int(unix_time)*cls.MAX_NANOSEC)

    @classmethod
    @deprecated(version="4.0.0",
                reason="This method is deprecated. TimeOffset has been merged into Timestamp.")
    def from_timeoffset(cls, toff: TimestampConstructionType) -> "Timestamp":
        """Legacy method that converted a TimeOffset to a Timestamp"""
        toff = mediatimestamp(toff)
        return cls(sec=toff.sec, ns=toff.ns, sign=toff.sign)

    @classmethod
    def get_interval_fraction(cls,
                              rate_num: RationalTypes,
                              rate_den: RationalTypes = 1,
                              factor: int = 1) -> "Timestamp":
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")
        if factor < 1:
            raise TsValueError("invalid interval factor")

        rate = Fraction(rate_num, rate_den)
        ns = int((cls.MAX_NANOSEC * rate.denominator) // (rate.numerator * factor))
        return cls(ns=ns)

    @classmethod
    def from_millisec(cls, millisec: int) -> "Timestamp":
        ns = millisec * 1000**2
        return cls(ns=ns)

    @classmethod
    def from_microsec(cls, microsec: int) -> "Timestamp":
        ns = microsec * 1000
        return cls(ns=ns)

    @classmethod
    def from_nanosec(cls, nanosec: int) -> "Timestamp":
        return cls(ns=nanosec)

    @classmethod
    def from_sec_frac(cls, toff_str: str) -> "Timestamp":
        sec_frac = toff_str.split(".")
        if len(sec_frac) != 1 and len(sec_frac) != 2:
            raise TsValueError("invalid second.fraction format")
        sec = int(sec_frac[0])
        sign = 1
        if sec_frac[0].startswith("-"):
            sign = -1
            sec = abs(sec)
        ns = 0
        if len(sec_frac) > 1:
            ns = _parse_seconds_fraction(sec_frac[1])
        return cls(sec=sec, ns=ns, sign=sign)

    @classmethod
    def from_tai_sec_frac(cls, ts_str: str) -> "Timestamp":
        return cls.from_sec_frac(ts_str)

    @classmethod
    def from_sec_nsec(cls, toff_str: str) -> "Timestamp":
        sec_frac = toff_str.split(":")
        if len(sec_frac) != 1 and len(sec_frac) != 2:
            raise TsValueError("invalid second:nanosecond format")
        sec = int(sec_frac[0])
        sign = 1
        if sec_frac[0].startswith("-"):
            sign = -1
            sec = abs(sec)
        ns = 0
        if len(sec_frac) > 1:
            ns = int(sec_frac[1])
        return cls(sec=sec, ns=ns, sign=sign)

    @classmethod
    def from_tai_sec_nsec(cls, ts_str: str) -> "Timestamp":
        return cls.from_sec_nsec(ts_str)

    @classmethod
    def from_float(cls, toff_float: float) -> "Timestamp":
        """Parse a float as a Timestamp
        """
        sign = 1
        if toff_float < 0:
            sign = -1
        ns = int(abs(toff_float) * cls.MAX_NANOSEC)
        return cls(ns=ns, sign=sign)

    @classmethod
    def from_datetime(cls, dt: datetime) -> "Timestamp":
        minTs = datetime.fromtimestamp(0, tz.gettz('UTC'))
        utcdt = dt.astimezone(tz.gettz('UTC'))
        seconds = int((utcdt - minTs).total_seconds())
        nanoseconds = utcdt.microsecond * 1000

        return cls.from_unix(seconds, nanoseconds, False)

    @classmethod
    def from_iso8601_utc(cls, iso8601utc: str) -> "Timestamp":
        if not iso8601utc.endswith('Z'):
            raise TsValueError("missing 'Z' at end of ISO 8601 UTC format")
        year, month, day, hour, minute, second, ns = _parse_iso8601(iso8601utc[:-1])
        gmtuple = (year, month, day, hour, minute, second - (second == 60))
        secs_since_epoch = calendar.timegm(gmtuple)
        return cls.from_unix(secs_since_epoch, ns, (second == 60))

    @classmethod
    def from_smpte_timelabel(cls, timelabel: str) -> "Timestamp":
        r = re.compile(r'(\d+)-(\d+)-(\d+)T(\d+):(\d+):(\d+)F(\d+) (\d+)/(\d+) UTC([-\+])(\d+):(\d+) TAI([-\+])(\d+)')
        m = r.match(timelabel)
        if m is None:
            raise TsValueError("invalid SMPTE Time Label string format")
        groups = m.groups()
        leap_sec = int(int(groups[5]) == 60)
        local_tm_sec = calendar.timegm(time.struct_time((int(groups[0]), int(groups[1]), int(groups[2]),
                                                         int(groups[3]), int(groups[4]), int(groups[5]) - leap_sec,
                                                         0, 0, 0)))
        rate_num = int(groups[7])
        rate_den = int(groups[8])
        utc_sign = 1
        if groups[9] == '-':
            utc_sign = -1
        tai_sign = 1
        if groups[12] == '-':
            tai_sign = -1
        tai_seconds = (local_tm_sec +
                       leap_sec -
                       utc_sign*(int(groups[10])*60*60 + int(groups[11])*60) -
                       tai_sign*int(groups[13]))
        count = Timestamp(tai_seconds, 0).to_count(rate_num, rate_den, cls.ROUND_UP)
        count += int(groups[6])
        return cls.from_count(count, rate_num, rate_den)

    @classmethod
    def from_str(cls, ts_str: str,) -> "Timestamp":
        """Parse a string as a TimeStamp

        Accepts:
        * SMPTE timelabel
        * ISO 8601 UTC
        * second:nanosecond
        * second.fraction formats
        * "now" to mean the current time.
        """
        if 'F' in ts_str:
            return cls.from_smpte_timelabel(ts_str)
        elif 'T' in ts_str:
            return cls.from_iso8601_utc(ts_str)
        elif ts_str.strip() == 'now':
            return cls.get_time()
        elif '.' in ts_str:
            return cls.from_sec_frac(ts_str)
        else:
            return cls.from_sec_nsec(ts_str)

    @classmethod
    def from_count(cls, count: int, rate_num: RationalTypes, rate_den: RationalTypes = 1) -> "Timestamp":
        """Returns a new Timestamp derived from a count and a particular rate.

        :param count: The sample count
        :param rate_num: The numerator of the rate, in Hz
        :param rate_den: The denominator of the rate in Hz
        """
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")
        sign = 1
        if count < 0:
            sign = -1
        rate = Fraction(rate_num, rate_den)
        ns = (cls.MAX_NANOSEC * abs(count) * rate.denominator) // rate.numerator
        return cls(ns=ns, sign=sign)

    @classmethod
    def from_unix(cls, unix_sec: int, unix_ns: int, is_leap: bool = False) -> "Timestamp":
        leap_sec = 0
        for tbl_sec, tbl_tai_sec_minus_1 in UTC_LEAP:
            if unix_sec >= tbl_sec:
                leap_sec = (tbl_tai_sec_minus_1 + 1) - tbl_sec
                break
        return cls(sec=unix_sec+leap_sec+is_leap, ns=unix_ns)

    @classmethod
    def from_utc(cls, utc_sec: int, utc_ns: int, is_leap: bool = False) -> "Timestamp":
        """ Wrapper of from_unix for back-compatibility.
        """
        return cls.from_unix(utc_sec, utc_ns, is_leap)

    def is_null(self) -> bool:
        return self._value == 0

    def get_leap_seconds(self) -> int:
        """ Get the UTC leaps seconds.
        Returns the number of leap seconds that the timestamp is adjusted by when
        converting to UTC.
        """
        leap_sec = 0
        for unix_sec, tai_sec_minus_1 in UTC_LEAP:
            if self.sec >= tai_sec_minus_1:
                leap_sec = (tai_sec_minus_1 + 1) - unix_sec
                break

        return leap_sec

    def to_millisec(self, rounding: "Timestamp.Rounding" = ROUND_NEAREST) -> int:
        use_rounding = rounding
        if self.sign < 0:
            if use_rounding == self.ROUND_UP:
                use_rounding = self.ROUND_DOWN
            elif use_rounding == self.ROUND_DOWN:
                use_rounding = self.ROUND_UP
        round_ns = 0
        if use_rounding == self.ROUND_NEAREST:
            round_ns = 1000**2 // 2
        elif use_rounding == self.ROUND_UP:
            round_ns = 1000**2 - 1
        return int(self.sign * ((abs(self._value) + round_ns) // 1000**2))

    def to_microsec(self, rounding: "Timestamp.Rounding" = ROUND_NEAREST) -> int:
        use_rounding = rounding
        if self.sign < 0:
            if use_rounding == self.ROUND_UP:
                use_rounding = self.ROUND_DOWN
            elif use_rounding == self.ROUND_DOWN:
                use_rounding = self.ROUND_UP
        round_ns = 0
        if use_rounding == self.ROUND_NEAREST:
            round_ns = 1000 // 2
        elif use_rounding == self.ROUND_UP:
            round_ns = 1000 - 1
        return int(self.sign * ((abs(self._value) + round_ns) // 1000))

    def to_nanosec(self) -> int:
        return self._value

    def to_sec_nsec(self) -> str:
        """ Convert to <seconds>:<nanoseconds>
        """
        strSign = ""
        if self.sign < 0:
            strSign = "-"
        return u"{}{}:{}".format(strSign, self.sec, self.ns)

    def to_sec_frac(self, fixed_size: bool = False) -> str:
        """ Convert to <seconds>.<fraction>
        """
        strSign = ""
        if self.sign < 0:
            strSign = "-"
        return u"{}{}.{}".format(
            strSign,
            self.sec,
            self._get_fractional_seconds(fixed_size=fixed_size))

    def to_tai_sec_nsec(self) -> str:
        return self.to_sec_nsec()

    def to_tai_sec_frac(self, fixed_size: bool = False) -> str:
        return self.to_sec_frac(fixed_size=fixed_size)

    def to_datetime(self) -> datetime:
        sec, nsec, leap = self.to_unix()
        microsecond = int(round(nsec/1000))
        if microsecond > 999999:
            sec += 1
            microsecond = 0
        dt = datetime.fromtimestamp(sec, tz.gettz('UTC'))
        dt = dt.replace(microsecond=microsecond)

        return dt

    def to_unix(self) -> Tuple[int, int, bool]:
        """ Convert to unix seconds.
        Returns a tuple of (seconds, nanoseconds, is_leap), where `is_leap` is
        `True` when the input time corresponds exactly to a UTC leap second.
        Note that this deliberately returns a tuple, to try and avoid confusion.
        """
        leap_sec = 0
        is_leap = False
        for unix_sec, tai_sec_minus_1 in UTC_LEAP:
            if self.sec >= tai_sec_minus_1:
                leap_sec = (tai_sec_minus_1 + 1) - unix_sec
                is_leap = self.sec == tai_sec_minus_1
                break

        return (self.sec - leap_sec, self.ns, is_leap)

    def to_utc(self) -> Tuple[int, int, bool]:
        """ Wrapper of to_unix for back-compatibility.
        """
        return self.to_unix()

    def to_iso8601_utc(self) -> str:
        """ Get printed representation in ISO8601 format (UTC)
        YYYY-MM-DDThh:mm:ss.s
        where `s` is fractional seconds at nanosecond precision (always 9-chars wide)
        """
        unix_s, unix_ns, is_leap = self.to_unix()
        utc_bd = time.gmtime(unix_s)
        frac_sec = self._get_fractional_seconds(fixed_size=True)
        leap_sec = int(is_leap)
        return '%04d-%02d-%02dT%02d:%02d:%02d.%sZ' % (utc_bd.tm_year,
                                                      utc_bd.tm_mon,
                                                      utc_bd.tm_mday,
                                                      utc_bd.tm_hour,
                                                      utc_bd.tm_min,
                                                      utc_bd.tm_sec + leap_sec,
                                                      frac_sec)

    def to_smpte_timelabel(self,
                           rate_num: RationalTypes,
                           rate_den: RationalTypes = 1,
                           utc_offset: Optional[int] = None) -> str:
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")
        rate = Fraction(rate_num, rate_den)
        count = self.to_count(rate)
        normalised_ts = Timestamp.from_count(count, rate)
        tai_seconds = normalised_ts.sec
        count_on_or_after_second = Timestamp(tai_seconds, 0).to_count(rate, rounding=self.ROUND_UP)
        count_within_second = count - count_on_or_after_second

        unix_sec, unix_ns, is_leap = normalised_ts.to_unix()
        leap_sec = int(is_leap)

        if utc_offset is None:
            # calculate local time offset
            utc_offset_sec = time.timezone
            lt = time.localtime(unix_sec)
            if lt.tm_isdst > 0:
                utc_offset_sec += 60*60
        else:
            utc_offset_sec = utc_offset
        utc_offset_sec_abs = abs(utc_offset_sec)
        utc_offset_hour = utc_offset_sec_abs // (60*60)
        utc_offset_min = (utc_offset_sec_abs % (60*60)) // 60
        utc_sign_char = '+'
        if utc_offset_sec < 0:
            utc_sign_char = '-'

        utc_bd = time.gmtime(unix_sec + utc_offset_sec)

        tai_offset = unix_sec + leap_sec - tai_seconds
        tai_sign_char = '+'
        if tai_offset < 0:
            tai_sign_char = '-'

        return '%04d-%02d-%02dT%02d:%02d:%02dF%02u %u/%u UTC%c%02u:%02u TAI%c%u' % (
                    utc_bd.tm_year, utc_bd.tm_mon, utc_bd.tm_mday,
                    utc_bd.tm_hour, utc_bd.tm_min, utc_bd.tm_sec + leap_sec,
                    count_within_second,
                    rate.numerator, rate.denominator,
                    utc_sign_char, utc_offset_hour, utc_offset_min,
                    tai_sign_char, abs(tai_offset))

    def to_count(self, rate_num: RationalTypes, rate_den: RationalTypes = 1,
                 rounding: "Timestamp.Rounding" = ROUND_NEAREST) -> int:
        """Returns an integer such that if this Timestamp is equal to an exact number of samples at the given rate
        then this is equal, and otherwise the value is rounded as indicated by the rounding parameter.

        :param rate_num: numerator of rate
        :param rate_den: denominator of rate
        :param rounding: One of Timestamp.ROUND_NEAREST, Timestamp.ROUND_UP, or Timestamp.ROUND_DOWN
        """
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")

        rate = Fraction(rate_num, rate_den)
        use_rounding = rounding
        if self.sign < 0:
            if use_rounding == self.ROUND_UP:
                use_rounding = self.ROUND_DOWN
            elif use_rounding == self.ROUND_DOWN:
                use_rounding = self.ROUND_UP
        if use_rounding == self.ROUND_NEAREST:
            round_ns = Timestamp.get_interval_fraction(rate, factor=2).to_nanosec()
        elif use_rounding == self.ROUND_UP:
            round_ns = Timestamp.get_interval_fraction(rate, factor=1).to_nanosec() - 1
        else:
            round_ns = 0

        return int(self.sign * (
                    ((abs(self._value) + round_ns) * rate.numerator) // (
                        self.MAX_NANOSEC * rate.denominator)))

    def to_phase_offset(self, rate_num: RationalTypes, rate_den: RationalTypes = 1) -> "Timestamp":
        """Return the smallest positive Timestamp such that abs(self - returnval) represents an integer number of
        samples at the given rate"""
        return self - self.normalise(rate_num, rate_den, rounding=Timestamp.ROUND_DOWN)

    def normalise(self,
                  rate_num: RationalTypes,
                  rate_den: RationalTypes = 1,
                  rounding: "Timestamp.Rounding" = ROUND_NEAREST) -> "Timestamp":
        """Return the nearest Timestamp to self which represents an integer number of samples at the given rate.

        :param rate_num: Rate numerator
        :param rate_den: Rate denominator
        :param rounding: How to round, if set to Timestamp.ROUND_DOWN (resp. Timestamp.ROUND_UP) this method will only
                         return a Timestamp less than or equal to this one (resp. greater than or equal to).
        """
        return self.from_count(self.to_count(rate_num, rate_den, rounding), rate_num, rate_den)

    def compare(self, other_in: TimestampConstructionType) -> int:
        other = mediatimestamp(other_in)
        if self._value > other._value:
            return 1
        elif self._value < other._value:
            return -1
        else:
            return 0

    def __str__(self) -> str:
        return self.to_sec_nsec()

    def __repr__(self) -> str:
        return "{}.from_sec_nsec({!r})".format("mediatimestamp.immutable." + type(self).__name__, self.to_sec_nsec())

    def __abs__(self) -> "Timestamp":
        return Timestamp(self.sec, self.ns, 1)

    def __hash__(self) -> int:
        return self.to_nanosec()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, (int, float, Timestamp)) and self.compare(other) == 0

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __lt__(self, other: TimestampConstructionType) -> bool:
        return self.compare(other) < 0

    def __le__(self, other: TimestampConstructionType) -> bool:
        return self.compare(other) <= 0

    def __gt__(self, other: TimestampConstructionType) -> bool:
        return self.compare(other) > 0

    def __ge__(self, other: TimestampConstructionType) -> bool:
        return self.compare(other) >= 0

    def __add__(self, other_in: TimestampConstructionType) -> "Timestamp":
        other = mediatimestamp(other_in)
        ns = self._value + other._value
        return Timestamp(ns=ns)

    def __sub__(self, other_in: TimestampConstructionType) -> "Timestamp":
        other = mediatimestamp(other_in)
        ns = self._value - other._value
        return Timestamp(ns=ns)

    def __iadd__(self, other_in: TimestampConstructionType) -> "Timestamp":
        other = mediatimestamp(other_in)
        tmp = self + other
        return self.__class__(ns=tmp._value)

    def __isub__(self, other_in: TimestampConstructionType) -> "Timestamp":
        other = mediatimestamp(other_in)
        tmp = self - other
        return self.__class__(ns=tmp._value)

    def __mul__(self, anint: int) -> "Timestamp":
        ns = self._value * anint
        return Timestamp(ns=ns)

    def __rmul__(self, anint: int) -> "Timestamp":
        return (self * anint)

    def __div__(self, anint: int) -> "Timestamp":
        return (self // anint)

    def __truediv__(self, anint: int) -> "Timestamp":
        return (self // anint)

    def __floordiv__(self, anint: int) -> "Timestamp":
        ns = self._value // anint
        return Timestamp(ns=ns)

    def _get_fractional_seconds(self, fixed_size: bool = False) -> str:
        div = self.MAX_NANOSEC // 10
        rem = self.ns
        sec_frac = ""

        for i in range(0, 9):
            if not fixed_size and i > 0 and rem == 0:
                break

            sec_frac += '%i' % (rem / div)
            rem %= div
            div //= 10

        return sec_frac



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# THESE CONSTANTS ARE NOT PART OF THIS LIBRARY'S PUBLIC INTERFACE
# The same values are made available by methods that are, such as
#
# Timestamp.MAX_NANOSEC
#
# So use those instead. At some point these constants could go away without warning

MAX_NANOSEC = 1000000000
MAX_SECONDS = 281474976710656

# The UTC leap seconds table below was extracted from the information provided at
# http://www.ietf.org/timezones/data/leap-seconds.list
#
# The order has been reversed.
# The NTP epoch seconds have been converted to Unix epoch seconds. The difference between
# the NTP epoch at 1 Jan 1900 and the Unix epoch at 1 Jan 1970 is 2208988800 seconds

UTC_LEAP = [
  # || UTC SEC  |  TAI SEC - 1 ||
  (1483228800, 1483228836),    # 1 Jan 2017, 37 leap seconds
  (1435708800, 1435708835),    # 1 Jul 2015, 36 leap seconds
  (1341100800, 1341100834),    # 1 Jul 2012, 35 leap seconds
  (1230768000, 1230768033),    # 1 Jan 2009, 34 leap seconds
  (1136073600, 1136073632),    # 1 Jan 2006, 33 leap seconds
  (915148800,  915148831),     # 1 Jan 1999, 32 leap seconds
  (867715200,  867715230),     # 1 Jul 1997, 31 leap seconds
  (820454400,  820454429),     # 1 Jan 1996, 30 leap seconds
  (773020800,  773020828),     # 1 Jul 1994, 29 leap seconds
  (741484800,  741484827),     # 1 Jul 1993, 28 leap seconds
  (709948800,  709948826),     # 1 Jul 1992, 27 leap seconds
  (662688000,  662688025),     # 1 Jan 1991, 26 leap seconds
  (631152000,  631152024),     # 1 Jan 1990, 25 leap seconds
  (567993600,  567993623),     # 1 Jan 1988, 24 leap seconds
  (489024000,  489024022),     # 1 Jul 1985, 23 leap seconds
  (425865600,  425865621),     # 1 Jul 1983, 22 leap seconds
  (394329600,  394329620),     # 1 Jul 1982, 21 leap seconds
  (362793600,  362793619),     # 1 Jul 1981, 20 leap seconds
  (315532800,  315532818),     # 1 Jan 1980, 19 leap seconds
  (283996800,  283996817),     # 1 Jan 1979, 18 leap seconds
  (252460800,  252460816),     # 1 Jan 1978, 17 leap seconds
  (220924800,  220924815),     # 1 Jan 1977, 16 leap seconds
  (189302400,  189302414),     # 1 Jan 1976, 15 leap seconds
  (157766400,  157766413),     # 1 Jan 1975, 14 leap seconds
  (126230400,  126230412),     # 1 Jan 1974, 13 leap seconds
  (94694400,   94694411),      # 1 Jan 1973, 12 leap seconds
  (78796800,   78796810),      # 1 Jul 1972, 11 leap seconds
  (63072000,   63072009),      # 1 Jan 1972, 10 leap seconds
]



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This library provides a class Timestamp which stores a signed time difference value with nanosecond precision.

The Timestamp can represent a time offset since the epoch (ie. 1970-01-01T00:00:00.000000000Z)

These data types are of use in a number of situations, but particularly for code that will handle PTP timestamps, which
are normally stored in this fashion.
"""

from .exceptions import TsValueError
from .immutable import (
    TimeOffset, SupportsMediaTimeOffset, mediatimeoffset,
    Timestamp, SupportsMediaTimestamp, mediatimestamp,
    TimeRange, SupportsMediaTimeRange, mediatimerange)

from .count_range import CountRange
from .time_value import TimeValue, TimeValueConstructTypes
from .time_value_range import TimeValueRange, RangeConstructionTypes, RangeTypes

__all__ = [
    "TsValueError",
    "TimeOffset", "SupportsMediaTimeOffset", "mediatimeoffset",
    "Timestamp", "SupportsMediaTimestamp", "mediatimestamp",
    "TimeRange", "SupportsMediaTimeRange", "mediatimerange",
    "CountRange",
    "TimeValue", "TimeValueConstructTypes",
    "TimeValueRange", "RangeConstructionTypes", "RangeTypes"]



#!/usr/bin/python
#
# Copyright 2018 British Broadcasting Corporation
#
# This is an internal BBC tool and is not licensed externally
# If you have received a copy of this erroneously then you do
# not have permission to reproduce it.

"""Hypothesis ( https://hypothesis.readthedocs.io/en/latest/ ) is a library used for generating data driven testing in
python. When used it requires "strategies", which are special data generators used by its engine to produce large
quantities of data to test hypothetical features of software.

This module provides additional strategies for hypothesis which generate the types of objects provided by this library,
it will be of use if using hypothesis to test code which depends upon mediatimestamp.
"""

from hypothesis.strategies import integers, lists

from mediatimestamp.immutable import Timestamp as ImmutableTimestamp
from mediatimestamp.immutable import TimeOffset as ImmutableTimeOffset
from mediatimestamp.immutable import TimeRange as ImmutableTimeRange


__all__ = ["timestamps", "timeoffsets", "timeranges", "disjoint_timeranges",
           "immutabletimestamps", "immutabletimeoffsets", "immutabletimeranges", "disjoint_immutabletimeranges"]

MIN_IMMUTABLETIMESTAMP = ImmutableTimestamp(ImmutableTimestamp.MAX_SECONDS, ImmutableTimestamp.MAX_NANOSEC - 1, -1)
MAX_IMMUTABLETIMESTAMP = ImmutableTimestamp(ImmutableTimestamp.MAX_SECONDS, ImmutableTimestamp.MAX_NANOSEC - 1)

MIN_IMMUTABLETIMEOFFSET = ImmutableTimeOffset(ImmutableTimeOffset.MAX_SECONDS, ImmutableTimeOffset.MAX_NANOSEC - 1, -1)
MAX_IMMUTABLETIMEOFFSET = ImmutableTimestamp(ImmutableTimeOffset.MAX_SECONDS, ImmutableTimeOffset.MAX_NANOSEC - 1)


#
# Immutable versions. It is recommended to use these where possible
#

def immutabletimestamps(min_value=MIN_IMMUTABLETIMESTAMP, max_value=MAX_IMMUTABLETIMESTAMP):
    """Draw from this strategy to get immutable timestamps between the given minimum and maximum values.
    Shrinks towards earlier timestamps."""

    # This is pretty straightforward: generates integers in the right range and interprets then as nanosecond values to
    # generate Timestamps with
    return integers(min_value=min_value.to_nanosec(),
                    max_value=max_value.to_nanosec()).map(ImmutableTimestamp.from_nanosec)


def immutabletimeoffsets(min_value=MIN_IMMUTABLETIMEOFFSET, max_value=MAX_IMMUTABLETIMEOFFSET):
    """Draw from this strategy to get immutable timeoffsets between the given minimum and maximum values.
    Shrinks towards zero."""

    # This is pretty straightforward: generates integers in the right range and interprets then as nanosecond values to
    # generate TimeOffsets with
    return integers(min_value=min_value.to_nanosec(),
                    max_value=max_value.to_nanosec()).map(ImmutableTimeOffset.from_nanosec)


def immutabletimeranges_of_length(length,
                                  in_range=ImmutableTimeRange.eternity(),
                                  inclusivity=ImmutableTimeRange.INCLUSIVE):
    min_value = in_range.start if in_range.start is not None else MIN_IMMUTABLETIMESTAMP
    max_value = (in_range.end if in_range.end is not None else MAX_IMMUTABLETIMESTAMP) - length

    # This is also straightforward in a way: generate a timestamp in the range and then create a timerange of length
    # starting at that timestamp
    return (immutabletimestamps(min_value=min_value, max_value=max_value)
            .map(lambda start: ImmutableTimeRange.from_start_length(start, length, inclusivity)))


def immutabletimeranges(in_range=ImmutableTimeRange.eternity(),
                        inclusivity=ImmutableTimeRange.INCLUSIVE):
    """Draw from this strategy to get non-empty immutable timeranges with the specified inclusivity completely contained
    in the specified range. Shrinks towards smaller ranges and earlier ones."""
    max_length = in_range.length if not isinstance(in_range.length, float) else MAX_IMMUTABLETIMESTAMP

    # This is a tad more complex: each time a range is drawn from this we generate a timestamp to use as a length and
    # then generate a timerange of that length
    return (immutabletimestamps(min_value=MIN_IMMUTABLETIMESTAMP, max_value=max_length)
            .flatmap(lambda k: immutabletimeranges_of_length(k, in_range=in_range, inclusivity=inclusivity)))


def disjoint_immutabletimeranges(in_range=ImmutableTimeRange.eternity(), min_size=0, max_size=None):
    """Draw from this strategy to get lists of non-overlapping immutable TimeRange classes all of which will be
    inclusive and contained within the specified range.

    Shrinks towards fewer ranges, smaller ones, and earlier ones."""

    min_value = in_range.start if in_range.start is not None else MIN_IMMUTABLETIMESTAMP
    max_value = in_range.end if in_range.end is not None else MAX_IMMUTABLETIMESTAMP

    # The most complex, probably. Each time an item is drawn generate an integer and use it as a list size,
    # then generate a list of that length of timestamps. Then sort them into ascending order. Then use each adjacent
    # pair as the start and end of a timerange. Finally reject any lists of timeranges where some timeranges overlap
    # (the construction ensures that the only way this can happen is if the overlap is a single timestamp)
    return (integers(min_value=min_size, max_value=max_size)
            .flatmap(lambda n: lists(
                immutabletimestamps(min_value=min_value, max_value=max_value),
                min_size=2*n,
                max_size=2*n
            )).map(sorted).map(
                lambda k: [ImmutableTimeRange(k[2*x+0], k[2*x+1], ImmutableTimeRange.INCLUSIVE)
                           for x in range(0, len(k)//2)]
            ).filter(
                lambda k: all(k[x].end != k[x+1].start for x in range(0, len(k) - 1))
            ))


timeoffsets = immutabletimeoffsets
timestamps = immutabletimestamps
timeranges = immutabletimeranges
disjoint_timeranges = disjoint_immutabletimeranges



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

# Basic metadata
name = 'mediatimestamp'
description = 'A timestamp library for high precision nanosecond timestamps'
url = 'https://github.com/bbc/rd-apmm-python-lib-mediatimestamp'
author = 'BBC R&D'
author_email = 'cloudfit-opensource@rd.bbc.co.uk'
license = 'Apache 2'
long_description = description


# Execute version file to set version variable
try:
    with open(("{}/_version.py".format(name)), "r") as fp:
        exec(fp.read())
except IOError:
    # Version file doesn't exist, fake it for now
    __version__ = "0.0.0"

package_names = [
    'mediatimestamp',
    'mediatimestamp.hypothesis',
    'mediatimestamp.immutable'
]
packages = {
    pkg: pkg.replace('.', '/') for pkg in package_names
}

# This is where you list packages which are required
packages_required = [
    "python-dateutil>=2.1",
    "deprecated"
]

setup(name=name,
      python_requires='>=3.10.0',
      version=__version__,
      description=description,
      url=url,
      author=author,
      author_email=author_email,
      license=license,
      packages=package_names,
      package_dir=packages,
      package_data={package_name: ['py.typed'] for package_name in package_names},
      install_requires=packages_required,
      scripts=[],
      data_files=[],
      long_description=long_description)
